class Singleton(type):

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        cls.instance = None

    def __call__(cls, *args, **kwargs):
        if not cls.instance:
            cls.instance = super().__call__(*args, **kwargs)

        return cls.instance





class XSS:

    def __init__(self):
        pass


class LFI:

    def __init__(self):
        pass


class RFI:

    def __init__(self):
        pass


class Traversal:

    def __init__(self):
        pass


class RCE:

    def __init__(self):
        pass


class OWASPScanner:

    def __init__(self):
        self.running_threads = []

        # params
        pass

    def discover_xss(self):
        # if potentially vulnerable, create instance and run scan in a new thread
        pass

    def discover_lfi(self):
        # if potentially vulnerable, create instance and run scan in a new thread
        pass

    def discover_rfi(self):
        # if potentially vulnerable, create instance and run scan in a new thread
        pass

    def run(self):
        # Perform all checks, create relevant threads
        pass



import time
import asyncio
import threading
import click
import os

from TheDigger_src.utils.coloring import COLOR, COLORED_COMBOS
from TheDigger_src.utils.exceptions import TheDiggerException, HostHandlerException
from TheDigger_src.utils.request_handler import RequestHandler
from TheDigger_src.utils.logger import SystemOutLogger
from TheDigger_src.utils.help_utils import HelpUtilities
from TheDigger_src.lib.fuzzer import URLFuzzer
from TheDigger_src.lib.host import Host
from TheDigger_src.lib.scanner import Scanner, NmapScan, NmapVulnersScan, VulnersScanner
from TheDigger_src.lib.sub_domain import SubDomainEnumerator
from TheDigger_src.lib.dns_handler import DNSHandler
from TheDigger_src.lib.waf import WAF
from TheDigger_src.lib.tls import TLSHandler
from TheDigger_src.lib.web_app import WebApplicationScanner

# Set path for relative access to builtin files.
MY_PATH = os.path.abspath(os.path.dirname(__file__))


def intro(logger):
    logger.info("""{}
 
    /$$$$$$$$ /$$                 /$$$$$$$\ /$$                                        
   |__  $$__/| $$                | $$__  $$|__/                                        
      | $$   | $$$$$$$   /$$$$$$ | $$  \ $$ /$$  /$$$$$$   /$$$$$$   /$$$$$$   /$$$$$$ 
      | $$   | $$__  $$ /$$__  $$| $$  | $$| $$ /$$__  $$ /$$__  $$ /$$__  $$ /$$__  $$
      | $$   | $$  \ $$| $$$$$$$$| $$  | $$| $$| $$  \ $$| $$  \ $$| $$$$$$$$| $$  \__/
      | $$   | $$  | $$| $$_____/| $$  | $$| $$| $$  | $$| $$  | $$| $$_____/| $$      
      | $$   | $$  | $$|  $$$$$$$| $$$$$$$/| $$|  $$$$$$$|  $$$$$$$|  $$$$$$$| $$      
      |__/   |__/  |__/ \_______/|_______/ |__/ \____  $$ \____  $$ \_______/|__/      
                                                /$$  \ $$ /$$  \ $$                    
                                               |  $$$$$$/|  $$$$$$/                    
                                                \______/  \______/                     

{}


https://github.com/Jistrokz/TheDigger
-------------------------------------------------------------------
    """.format(COLOR.RED, COLOR.RESET))


@click.command()
@click.version_option("0.8.5")
@click.argument("target")
@click.option("-d", "--dns-records", default="A,MX,NS,CNAME,SOA,TXT",
              help="Comma separated DNS records to query. Defaults to: A,MX,NS,CNAME,SOA,TXT")
@click.option("--tor-routing", is_flag=True, help="Route HTTP traffic through Tor (uses port 9050)."
                                                  " Slows total runtime significantly")
@click.option("--proxy-list", help="Path to proxy list file that would be used for routing HTTP traffic."
                                   " A proxy from the list will be chosen at random for each request."
                                   " Slows total runtime")
@click.option("-c", "--cookies", help="Comma separated cookies to add to the requests. "
                                      "Should be in the form of key:value\n"
                                      "Example: PHPSESSID:12345,isMobile:false")
@click.option("--proxy", help="Proxy address to route HTTP traffic through. Slows total runtime")
@click.option("-w", "--wordlist", default=os.path.join(MY_PATH, "wordlists/fuzzlist"),
              help="Path to wordlist that would be used for URL fuzzing")
@click.option("-T", "--threads", default=25,
              help="Number of threads to use for URL Fuzzing/Subdomain enumeration. Default: 25")
@click.option("--ignored-response-codes", default="302,400,401,402,403,404,503,504",
              help="Comma separated list of HTTP status code to ignore for fuzzing."
                   " Defaults to: 302,400,401,402,403,404,503,504")
@click.option("--subdomain-list", default=os.path.join(MY_PATH, "wordlists/subdomains"),
              help="Path to subdomain list file that would be used for enumeration")
@click.option("-sc", "--scripts", is_flag=True, help="Run Nmap scan with -sC flag")
@click.option("-sv", "--services", is_flag=True, help="Run Nmap scan with -sV flag")
@click.option("-f", "--full-scan", is_flag=True, help="Run Nmap scan with both -sV and -sC")
@click.option("-p", "--port", help="Use this port range for Nmap scan instead of the default")
@click.option("--vulners-nmap-scan", is_flag=True, help="Perform an NmapVulners scan. "
                                                        "Runs instead of the regular Nmap scan and is longer.")
@click.option("--vulners-path", default=os.path.join(MY_PATH, "utils/misc/vulners.nse"),
              help="Path to the custom nmap_vulners.nse script."
                   "If not used, TheDigger uses the built-in script it ships with.")
@click.option("-fr", "--follow-redirects", is_flag=True, default=False,
              help="Follow redirects when fuzzing. Default: False (will not follow redirects)")
@click.option("--tls-port", default=443, help="Use this port for TLS queries. Default: 443")
@click.option("--skip-health-check", is_flag=True, help="Do not test for target host availability")
@click.option("--no-url-fuzzing", is_flag=True, help="Do not fuzz URLs")
@click.option("--no-sub-enum", is_flag=True, help="Do not bruteforce subdomains")
@click.option("--skip-nmap-scan", is_flag=True, help="Do not perform an Nmap scan")
#This is included for the future. This feature can be turned on to send request at intervals
# @click.option("-d", "--delay", default="0.25-1",
#               help="Min and Max number of seconds of delay to be waited between requests\n"
#                    "Defaults to Min: 0.25, Max: 1. Specified in the format of Min-Max")
@click.option("-q", "--quiet", is_flag=True, help="Do not output to stdout")
@click.option("-o", "--outdir", default="TheDigger_scan_results",
              help="Directory destination for scan output")
def main(target,
         tor_routing,
         proxy_list,
         proxy,
         cookies,
         dns_records,
         wordlist,
         threads,
         ignored_response_codes,
         subdomain_list,
         full_scan,
         scripts,
         services,
         port,
         vulners_nmap_scan,
         vulners_path,
         tls_port,
         skip_health_check,
         follow_redirects,
         no_url_fuzzing,
         no_sub_enum,
         skip_nmap_scan,
         # delay,
         outdir,
         quiet):
    try:
        # ------ Arg validation ------
        # Set logging level and Logger instance
        log_level = HelpUtilities.determine_verbosity(quiet)
        logger = SystemOutLogger(log_level)
        intro(logger)

        target = target.lower()
        try:
            HelpUtilities.validate_executables()
        except TheDiggerException as e:
            logger.critical(str(e))
            exit(9)
        HelpUtilities.validate_wordlist_args(proxy_list, wordlist, subdomain_list)
        HelpUtilities.validate_proxy_args(tor_routing, proxy, proxy_list)
        HelpUtilities.create_output_directory(outdir)

        if tor_routing:
            logger.info("{} Testing that Tor service is up...".format(COLORED_COMBOS.NOTIFY))
        elif proxy_list:
            if proxy_list and not os.path.isfile(proxy_list):
                raise FileNotFoundError("Not a valid file path, {}".format(proxy_list))
            else:
                logger.info("{} Routing traffic using proxies from list {}\n".format(
                    COLORED_COMBOS.NOTIFY, proxy_list))
        elif proxy:
            logger.info("{} Routing traffic through proxy {}\n".format(COLORED_COMBOS.NOTIFY, proxy))

       

        dns_records = tuple(dns_records.split(","))
        ignored_response_codes = tuple(int(code) for code in ignored_response_codes.split(","))

        if port:
            HelpUtilities.validate_port_range(port)

        # ------ /Arg validation ends------

        if cookies:
            try:
                cookies = HelpUtilities.parse_cookie_arg(cookies)
            except TheDiggerException as e:
                logger.critical("{}{}{}".format(COLOR.RED, str(e), COLOR.RESET))
                exit(2)

        # Set Request Handler instance
        request_handler = RequestHandler(
            proxy_list=proxy_list,
            tor_routing=tor_routing,
            single_proxy=proxy,
            cookies=cookies
        )

        if tor_routing:
            try:
                HelpUtilities.confirm_traffic_routs_through_tor()
                logger.info("{} Validated Tor service is up. Routing traffic anonymously\n".format(
                    COLORED_COMBOS.NOTIFY))
            except TheDiggerException as err:
                print("{}{}{}".format(COLOR.RED, str(err), COLOR.RESET))
                exit(3)

        main_loop = asyncio.get_event_loop()

        logger.info("{}### TheDigger has started Digging ###{}\n".format(COLOR.GRAY, COLOR.RESET))
        logger.info("{} Trying to gather information about host: {}".format(COLORED_COMBOS.INFO, target))

        # TODO: Populate array when multiple targets are supported
        # hosts = []
        try:
            host = Host(target=target, dns_records=dns_records)
            host.parse()
        except HostHandlerException as e:
            logger.critical("{}{}{}".format(COLOR.RED, str(e), COLOR.RESET))
            exit(11)

        if not skip_health_check:
            try:
                HelpUtilities.validate_target_is_up(host)
            except TheDiggerException as err:
                logger.critical("{}{}{}".format(COLOR.RED, str(err), COLOR.RESET))
                exit(42)

        if not skip_nmap_scan:
            if vulners_nmap_scan:
                logger.info("\n{} Setting NmapVulners scan to run in the background".format(COLORED_COMBOS.INFO))
                nmap_vulners_scan = NmapVulnersScan(host=host, port_range=port, vulners_path=vulners_path)
                nmap_thread = threading.Thread(target=VulnersScanner.run, args=(nmap_vulners_scan,))
                # Run NmapVulners scan in the background
                nmap_thread.start()
            else:
                logger.info("\n{} Setting Nmap scan to run in the background".format(COLORED_COMBOS.INFO))
                nmap_scan = NmapScan(
                    host=host,
                    port_range=port,
                    full_scan=full_scan,
                    scripts=scripts,
                    services=services)

                nmap_thread = threading.Thread(target=Scanner.run, args=(nmap_scan,))
                # Run Nmap scan in the background. Can take some time
                nmap_thread.start()

        # Run first set of checks - TLS, Web/WAF Data, DNS data
        waf = WAF(host)
        tls_info_scanner = TLSHandler(host, tls_port)
        web_app_scanner = WebApplicationScanner(host)
        tasks = (
            asyncio.ensure_future(tls_info_scanner.run()),
            asyncio.ensure_future(waf.detect()),
            asyncio.ensure_future(DNSHandler.grab_whois(host)),
            asyncio.ensure_future(web_app_scanner.run_scan()),
            asyncio.ensure_future(DNSHandler.generate_dns_dumpster_mapping(host, logger))
        )

        main_loop.run_until_complete(asyncio.wait(tasks))

        # Second set of checks - URL fuzzing, Subdomain enumeration
        if not no_url_fuzzing:
            fuzzer = URLFuzzer(host, ignored_response_codes, threads, wordlist, follow_redirects)
            main_loop.run_until_complete(fuzzer.fuzz_all())

        if not host.is_ip:
            sans = tls_info_scanner.sni_data.get("SANs")
            subdomain_enumerator = SubDomainEnumerator(
                host,
                domain_list=subdomain_list,
                sans=sans,
                ignored_response_codes=ignored_response_codes,
                num_threads=threads,
                follow_redirects=follow_redirects,
                no_sub_enum=no_sub_enum
            )
            main_loop.run_until_complete(subdomain_enumerator.run())

        if not skip_nmap_scan:
            if nmap_thread.is_alive():
                logger.info("{} All scans done. Waiting for Nmap scan to wrap up. "
                            "Time left may vary depending on scan type and port range".format(COLORED_COMBOS.INFO))

                while nmap_thread.is_alive():
                    time.sleep(15)

        logger.info("\n{}### TheDigger has reached the surface. Digging has stopped ###{}\n".format(COLOR.GRAY, COLOR.RESET))
        os.system("stty sane")

    except KeyboardInterrupt:
        print("{}Keyboard Interrupt detected. Exiting{}".format(COLOR.RED, COLOR.RESET))
        # terminal after CTRL+C
        os.system("stty sane")
        exit(42)


if __name__ == "__main__":
    main()



from collections import namedtuple

Color = namedtuple("Color", ["RED", "BLUE", "CYAN", "GREEN", "YELLOW", "GRAY", "BOLD", "RESET"])
COLOR = Color(
    "\033[1;31m",  # red
    "\033[1;34m",  # blue
    "\033[1;36m",  # cyan
    "\033[1;32m",  # green
    "\033[93m",    # yellow
    "\033[1;30m",  # gray
    "\033[;1m",    # bold
    "\033[0;0m"    # reset
)

ColoredCombos = namedtuple("ColoredCombos", ["INFO", "GOOD", "BAD", "NOTIFY"])
COLORED_COMBOS = ColoredCombos(
    "{}[#]{}".format(COLOR.BLUE, COLOR.RESET),
    "{}[v]{}".format(COLOR.GREEN, COLOR.RESET),
    "{}[x]{}".format(COLOR.RED, COLOR.RESET),
    "{}[!]{}".format(COLOR.YELLOW, COLOR.RESET))



from dns import resolver
from asyncio.subprocess import PIPE, create_subprocess_exec
from requests.exceptions import ConnectionError
from TheDigger_src.utils.help_utils import HelpUtilities
from TheDigger_src.utils.exceptions import TheDiggerException
from TheDigger_src.utils.logger import Logger
from TheDigger_src.utils.coloring import COLOR, COLORED_COMBOS


# noinspection PyUnboundLocalVariable
class DNSHandler:
    """Handles DNS queries and lookups"""

    resolver = resolver.Resolver()

    @classmethod
    def query_dns(cls, domains, records):
        """
        Query DNS records for host.
        :param domains: Iterable of domains to get DNS Records for
        :param records: Iterable of DNS records to get from domain.
        """
        results = {k: set() for k in records}
        for record in records:
            for domain in domains:
                try:
                    answers = cls.resolver.query(domain, record)
                    for answer in answers:
                        # Add value to record type
                        results.get(record).add(answer)
                except (resolver.NoAnswer, resolver.NXDOMAIN, resolver.NoNameservers):
                    # Type of record doesn't fit domain or no answer from ns
                    continue

        return {k: v for k, v in results.items() if v}

    @classmethod
    async def grab_whois(cls, host):
        if not host.naked:
            return

        script = "whois {}".format(host.naked).split()
        log_file = HelpUtilities.get_output_path("{}/whois.txt".format(host.target))
        logger = Logger(log_file)

        process = await create_subprocess_exec(
            *script,
            stdout=PIPE,
            stderr=PIPE
        )
        result, err = await process.communicate() #err has not been used, Please make sure to implement the variable when an er

        if process.returncode == 0:
            logger.info("{} {} WHOIS information retrieved".format(COLORED_COMBOS.GOOD, host))
            for line in result.decode().strip().split("\n"):
                    if ":" in line:
                        logger.debug(line)

    @classmethod
    async def generate_dns_dumpster_mapping(cls, host, sout_logger):
        sout_logger.info("{} Trying to fetch DNS Mapping for {} from DNS dumpster".format(
            COLORED_COMBOS.INFO, host))
        try:
            page = HelpUtilities.query_dns_dumpster(host=host)
            if page.status_code == 200:
                path = HelpUtilities.get_output_path("{}/dns_mapping.png".format(host.target))
                with open(path, "wb") as target_image:
                    target_image.write(page.content)
                sout_logger.info("{} Successfully fetched DNS mapping for {}".format(
                    COLORED_COMBOS.GOOD, host.target)
                )
            else:
                raise TheDiggerException
        except TheDiggerException:
            sout_logger.info("{} Failed to generate DNS mapping. A connection error occurred.".format(
                COLORED_COMBOS.BAD))



import os
import distutils.spawn
from platform import system
from collections import Counter
from subprocess import PIPE, check_call, CalledProcessError
from requests.exceptions import ConnectionError
from TheDigger_src.utils.exceptions import TheDiggerException, ScannerException, RequestHandlerException
from TheDigger_src.utils.request_handler import RequestHandler


class HelpUtilities:

    PATH = ""

    @classmethod
    def validate_target_is_up(cls, host):
        cmd = "ping -c 1 {}".format(host.target)
        try:
            check_call(cmd.split(), stdout=PIPE, stderr=PIPE)
            return
        except CalledProcessError:
            # Maybe ICMP is blocked. Try web server
            try:
                if host.port == 443 or host.port == 80:
                    url = "{}://{}".format(host.protocol, host.target)
                else:
                    url = "{}://{}:{}".format(host.protocol, host.target, host.port)
                rh = RequestHandler()
                rh.send("GET", url=url, timeout=15)
                return
            except (ConnectionError, RequestHandlerException):
                raise TheDiggerException("Target {} seems to be down (no response to ping or from a web server"
                                       " at port {}).\nRun with --skip-health-check to ignore hosts"
                                       " considered as down.".format(host, host.port))

    @classmethod
    def parse_cookie_arg(cls, cookie_arg):
        try:
            cookies = {}
            for c in cookie_arg.split(','):
                c = c.split(":")
                cookies[c[0]] = c[1]
            return cookies
        except (IndexError, TypeError):
            raise TheDiggerException("Cookie parsing error occurred, probably due to invalid cookie format.\n"
                                   "Cookie format should be comma separated key:value pairs. Use --help "
                                   "for more info.")

    @classmethod
    def validate_wordlist_args(cls, proxy_list, wordlist, subdomain_list):
        if proxy_list and not os.path.isfile(proxy_list):
            raise FileNotFoundError("Not a valid file path, {}".format(proxy_list))

        if wordlist and not os.path.isfile(wordlist):
            raise FileNotFoundError("Not a valid file path, {}".format(wordlist))

        if subdomain_list and not os.path.isfile(subdomain_list):
            raise FileNotFoundError("Not a valid file path, {}".format(wordlist))

    @classmethod
    def validate_port_range(cls, port_range):
        """Validate port range for Nmap scan"""
        ports = port_range.split("-")
        if all(ports) and int(ports[-1]) <= 65535 and not len(ports) != 2:
            return True
        raise ScannerException("Invalid port range {}".format(port_range))

    @classmethod
    def validate_proxy_args(cls, *args):
        """No more than 1 of the following can be specified: tor_routing, proxy, proxy_list"""
        supplied_proxies = Counter((not arg for arg in (*args,))).get(False)
        if not supplied_proxies:
            return
        elif supplied_proxies > 1:
            raise TheDiggerException("Must specify only one of the following:\n"
                                   "--tor-routing, --proxy-list, --proxy")

    @classmethod
    def determine_verbosity(cls, quiet):
        if quiet:
            return "CRITICAL"
        else:
            return "INFO"

    @classmethod
    def find_nmap_executable(cls):
        return distutils.spawn.find_executable("nmap")

    @classmethod
    def find_openssl_executable(cls):
        return distutils.spawn.find_executable("openssl")

    @classmethod
    def find_mac_gtimeout_executable(cls):
        """To add macOS support, the coreutils package needs to be installed using homebrew"""
        return distutils.spawn.find_executable("gtimeout")

    @classmethod
    def validate_executables(cls):
        if not (cls.find_nmap_executable() and cls.find_openssl_executable()):
            raise TheDiggerException("Could not find Nmap or OpenSSL "
                                   "installed. Please install them and run TheDigger again.")
        if system() == "Darwin":
            if not cls.find_mac_gtimeout_executable():
                raise TheDiggerException("To support TheDigger with macOS 'gtimeout' must be installed.\n"
                                       "gtimeout can be installed by running 'brew install coreutils'")
        return

    @classmethod
    def create_output_directory(cls, outdir):
        """Tries to create base output directory"""
        cls.PATH = outdir
        try:
            os.mkdir(outdir)
        except FileExistsError:
            pass

    @classmethod
    def get_output_path(cls, module_path):
        return "{}/{}".format(cls.PATH, module_path)

    @classmethod
    def confirm_traffic_routs_through_tor(cls):
        rh = RequestHandler()
        try:
            page = rh.send("GET", url="https://check.torproject.org")
            if "Congratulations. This browser is configured to use Tor." in page.text:
                return
            elif "Sorry. You are not using Tor" in page.text:
                raise TheDiggerException("Traffic does not seem to be routed through Tor.\nExiting")
        except RequestHandlerException:
            raise TheDiggerException("Tor service seems to be down - not able to connect to 127.0.0.1:9050.\nExiting")

    @classmethod
    def query_dns_dumpster(cls, host):
        # Start DNS Dumpster session for the token
        request_handler = RequestHandler()
        dnsdumpster_session = request_handler.get_new_session()
        url = "https://dnsdumpster.com"
        if host.naked:
            target = host.naked
        else:
            target = host.target
        payload = {
            "targetip": target,
            "csrfmiddlewaretoken": None
        }
        try:
            dnsdumpster_session.get(url, timeout=10)
            jar = dnsdumpster_session.cookies
            for c in jar:
                if not c.__dict__.get("name") == "csrftoken":
                    continue
                payload["csrfmiddlewaretoken"] = c.__dict__.get("value")
                break
            dnsdumpster_session.post(url, data=payload, headers={"Referer": "https://dnsdumpster.com/"})

            return dnsdumpster_session.get("https://dnsdumpster.com/static/map/{}.png".format(target))
        except ConnectionError:
            raise TheDiggerException

    @classmethod
    def extract_hosts_from_cidr(cls):
        pass

    @classmethod
    def extract_hosts_from_range(cls):
        pass



import uuid
from functools import partial
from multiprocessing.pool import ThreadPool
from requests.exceptions import ConnectionError
from TheDigger_src.utils.exceptions import FuzzerException, RequestHandlerException
from TheDigger_src.utils.coloring import COLOR, COLORED_COMBOS
from TheDigger_src.utils.request_handler import RequestHandler
from TheDigger_src.utils.help_utils import HelpUtilities
from TheDigger_src.utils.logger import Logger


# Really wanted to use Aiohttp, doesn't play nice with proxies or TOR, disconnects unexpectedly, etc.
# Going threaded on this one


class URLFuzzer:

    def __init__(self,
                 host,
                 ignored_response_codes,
                 num_threads,
                 path_to_wordlist,
                 follow_redirects=False):

        self.target = host.target
        self.ignored_error_codes = ignored_response_codes
        self.proto = host.protocol
        self.port = host.port
        self.num_threads = num_threads
        self.path_to_wordlist = path_to_wordlist
        self.wordlist = self._create_set_from_wordlist_file(path_to_wordlist)
        self.follow_redirects = follow_redirects
        self.request_handler = RequestHandler()  # Will get the single, already initiated instance
        self.logger = None

    @staticmethod
    def _create_set_from_wordlist_file(wordlist):
        try:
            with open(wordlist, "r") as file:
                fuzzlist = file.readlines()
                fuzzlist = [x.replace("\n", "") for x in fuzzlist]
                return set(fuzzlist)
        except FileNotFoundError:
            raise FuzzerException("Cannot open file {}. Will not perform Fuzzing".format(wordlist))

    def _log_response(self, code, url, headers):
        if 300 > code >= 200:
            color = COLOR.GREEN
        elif 400 > code >= 300:
            color = COLOR.BLUE
            url += " redirects to {}".format(headers.get("Location"))
        elif 510 > code >= 400:
            color = COLOR.RED
        else:
            color = COLOR.RESET
        self.logger.info("\t{}[{}]{} {}".format(
            color, code, COLOR.RESET, url))

    def _build_request_url(self, uri, sub_domain):
        if not sub_domain:
            if self.port != 80 and self.port != 443:
                url = "{}://{}:{}/{}".format(self.proto, self.target, self.port, uri)
            else:
                url = "{}://{}/{}".format(self.proto, self.target, uri)
        else:
            if self.port != 80 and self.port != 443:
                url = "{}://{}.{}:{}".format(self.proto, uri, self.target, self.port)
            else:
                url = "{}://{}.{}".format(self.proto, uri, self.target)
        return url

    def _fetch(self, uri, sub_domain=False):
        """
        Send a HEAD request to URL and print response code if it's not in ignored_error_codes
        :param uri: URI to fuzz
        :param sub_domain: If True, build destination URL with {URL}.{HOST} else {HOST}/{URL}
        """
        url = self._build_request_url(uri, sub_domain=sub_domain)

        try:
            res = self.request_handler.send("HEAD", url=url, allow_redirects=self.follow_redirects)
            if res.status_code not in self.ignored_error_codes:
                self._log_response(res.status_code, url, res.headers)
        except (AttributeError, RequestHandlerException):
            # res is None or another error occurred
            pass

    def get_log_file_path(self, path):
        if path:
            log_file = path
        else:
            log_file = "{}/url_fuzz.txt".format(self.target)

        return Logger(HelpUtilities.get_output_path(log_file))

    @staticmethod
    def _rule_out_false_positives(response_codes, sub_domain):
        if any(code == 200 for code in response_codes):
            if sub_domain:
                err_msg = "Wildcard subdomain support detected (all subdomains return 200)." \
                          " Will not bruteforce subdomains"
            else:
                err_msg = "Web server seems to redirect requests for all resources " \
                          "to eventually return 200. Will not bruteforce URLs"
            raise FuzzerException(err_msg)

    def _generate_fake_requests(self, sub_domain):
        response_codes = []
        fake_uris = (uuid.uuid4(), uuid.uuid4())
        session = self.request_handler.get_new_session()
        for uri in fake_uris:
            url = self._build_request_url(uri, sub_domain)
            try:
                res = self.request_handler.send("GET", url=url, allow_redirects=True)
                response_codes.append(res.status_code)
                res = session.get(url=url, allow_redirects=self.follow_redirects)
                response_codes.append(res.status_code)
            except RequestHandlerException as e:
                if sub_domain:  # If should-not-work.example.com doesn't resolve, no wildcard subdomain is present
                    return [0]
                else:
                    raise FuzzerException("Could not get a response from {}."
                                          " Maybe target is down ?".format(self.target))
        return response_codes

    async def fuzz_all(self, sub_domain=False, log_file_path=None):
        """
        Create a pool of threads and exhaust self.wordlist on self._fetch
        Should be run in an event loop.
        :param sub_domain: Indicate if this is subdomain enumeration or URL busting
        :param log_file_path: Log subdomain enum results to this path.
        """
        self.logger = self.get_log_file_path(log_file_path)
        try:
            # Rule out wildcard subdomain support/all resources redirect to a 200 page
            response_codes = self._generate_fake_requests(sub_domain)
            self._rule_out_false_positives(response_codes, sub_domain)

            if not sub_domain:
                self.logger.info("{} Fuzzing URLs".format(COLORED_COMBOS.INFO))
            self.logger.info("{} Reading from list: {}".format(COLORED_COMBOS.INFO, self.path_to_wordlist))
            pool = ThreadPool(self.num_threads)
            pool.map(partial(self._fetch, sub_domain=sub_domain), self.wordlist)
            pool.close()
            pool.join()
            if not sub_domain:
                self.logger.info("{} Done fuzzing URLs".format(COLORED_COMBOS.INFO))
        except FuzzerException as e:
            self.logger.info("{} {}".format(COLORED_COMBOS.BAD, e))
        except ConnectionError as e:
            if "Remote end closed connection without response" in str(e):
                self.logger.info("{} {}. Target is actively closing connections - will not "
                                 "bruteforce URLs".format(COLORED_COMBOS.BAD, str(e)))



from requests.exceptions import TooManyRedirects, ConnectionError
from TheDigger_src.utils.web_server_validator import WebServerValidator
from TheDigger_src.utils.exceptions import WAFException, WebServerValidatorException
from TheDigger_src.utils.request_handler import RequestHandler
from TheDigger_src.utils.coloring import COLOR, COLORED_COMBOS
from TheDigger_src.utils.help_utils import HelpUtilities
from TheDigger_src.utils.logger import Logger


SERVER = "Server"


class WAFApplicationMethods:

    @classmethod
    def detect_cloudfront(cls, res):
        service = "CloudFront"
        waf_headers = ("Via", "X-cache")
        if any(h in res.headers.keys() for h in waf_headers) and any(service.lower() in val for val in res.headers.values()):
            return True
        if res.headers.get(SERVER) == service:
            return True
        return

    @classmethod
    def detect_incapsula(cls, res):
        if "X-Iinfo" in res.headers.keys() or res.headers.get("X-CDN") == "Incapsula":
            return True
        return

    @classmethod
    def detect_distil(cls, res):
        if res.headers.get("x-distil-cs"):
            return True
        return

    @classmethod
    def detect_cloudflare(cls, res):
        if "CF-RAY" in res.headers.keys() or res.headers.get(SERVER) == "cloudflare":
            return True
        return

    @classmethod
    def detect_edgecast(cls, res):
        if SERVER in res.headers.keys() and "ECD" in res.headers[SERVER]:
            return True
        return

    @classmethod
    def detect_maxcdn(cls, res):
        if SERVER in res.headers.keys() and "NetDNA-cache" in res.headers[SERVER]:
            return True
        return

    @classmethod
    def detect_sucuri(cls, res):
        if any((
                res.headers.get(SERVER) == "Sucuri/Cloudproxy",
                "X-Sucuri-ID" in res.headers.keys(),
                "X-Sucuri-Cache"in res.headers.keys(),
                "Access Denied - Sucuri Website Firewall" in res.text)):
            return True
        return

    @classmethod
    def detect_reblaze(cls, res):
        if res.headers.get(SERVER) == "Reblaze Secure Web Gateway" or res.cookies.get("rbzid"):
            return True
        return


class WAF:

    def __init__(self, host):
        self.host = host
        self.cnames = host.dns_results.get('CNAME')
        self.request_handler = RequestHandler()
        self.web_server_validator = WebServerValidator()
        self.waf_present = False
        self.waf_cname_map = {
            "incapdns": "Incapsula",
            "edgekey": "Akamai",
            "akamai": "Akamai",
            "edgesuite": "Akamai",
            "distil": "Distil Networks",
            "cloudfront": "CloudFront",
            "netdna-cdn": "MaxCDN"
        }
        self.waf_app_method_map = {
            "CloudFront": WAFApplicationMethods.detect_cloudfront,
            "Cloudflare": WAFApplicationMethods.detect_cloudflare,
            "Incapsula": WAFApplicationMethods.detect_incapsula,
            "MaxCDN": WAFApplicationMethods.detect_maxcdn,
            "Edgecast": WAFApplicationMethods.detect_edgecast,
            "Distil Networks": WAFApplicationMethods.detect_distil,
            "Sucuri": WAFApplicationMethods.detect_sucuri,
            "Reblaze": WAFApplicationMethods.detect_reblaze
        }
        log_file = HelpUtilities.get_output_path("{}/WAF.txt".format(self.host.target))
        self.logger = Logger(log_file)

    def _waf_detected(self, name, where):
        self.logger.info(
            "{} Detected WAF presence in {}: {}{}{}".format(
                COLORED_COMBOS.BAD, where, COLOR.RED, name, COLOR.RESET))
        self.waf_present = True

    def _detect_by_cname(self):
        for waf in self.waf_cname_map:
            if any(waf in str(cname) for cname in self.cnames):
                self._waf_detected(self.waf_cname_map.get(waf), "CNAME record")

    async def _detect_by_application(self):
        try:
            session = self.request_handler.get_new_session()
            response = session.get(
                timeout=20,
                allow_redirects=True,
                url="{}://{}:{}".format(
                    self.host.protocol,
                    self.host.target,
                    self.host.port
                )
            )
            for waf, method in self.waf_app_method_map.items():
                result = method(response)
                if result:
                    self._waf_detected(waf, "web application")

        except (ConnectionError, TooManyRedirects) as e:
            raise WAFException("Couldn't get response from server.\n"
                               "Caused due to exception: {}".format(str(e)))

    async def detect(self):
        self.logger.info("{} Trying to detect WAF presence in {}".format(COLORED_COMBOS.INFO, self.host))
        if self.cnames:
            self._detect_by_cname()
        try:
            self.web_server_validator.validate_target_webserver(self.host)
            await self._detect_by_application()

            if not self.waf_present:
                self.logger.info("{} Did not detect WAF presence in target".format(COLORED_COMBOS.GOOD))
        except WebServerValidatorException:
            self.logger.info(
                "{} Target does not seem to have an active web server on port {}. "
                "No WAF could be detected on an application level.".format(COLORED_COMBOS.NOTIFY, self.host.port))


class TheDiggerException(Exception):
    """TheDigger base exception class"""
    def __init__(self, message='TheDigger Base Exception'):
        self._message = message

    def __str__(self):
        return self._message


class FuzzerException(TheDiggerException):
    def __init__(self, message='Fuzzer Exception'):
        super().__init__(message)

    def __str__(self):
        return self._message


class HostHandlerException(TheDiggerException):
    def __init__(self, message='Host Handler Exception'):
        super().__init__(message)

    def __str__(self):
        return self._message


class ScannerException(TheDiggerException):
    def __init__(self, message='Scanner Exception'):
        super().__init__(message)

    def __str__(self):
        return self._message


class WAFException(TheDiggerException):
    def __init__(self, message='WAF Exception'):
        super().__init__(message)

    def __str__(self):
        return self._message


class RequestHandlerException(TheDiggerException):

    def __init__(self, message='RequestHandler Exception'):
        super().__init__(message)

    def __str__(self):
        return self._message


class RequestHandlerConnectionReset(RequestHandlerException):

    def __init__(self, message='Connection Reset'):
        super().__init__(message)

    def __str__(self):
        return self._message


class WebAppScannerException(TheDiggerException):
    def __init__(self, message='Web Application Scanner Exception'):
        super().__init__(message)

    def __str__(self):
        return self._message


class WebServerValidatorException(TheDiggerException):
    def __init__(self, message='Web Server Validator Exception'):
        super().__init__(message)

    def __str__(self):
        return self._message



import random
from fake_useragent import UserAgent
from requests import request, Session, utils as requests_utils
from requests.exceptions import ProxyError, TooManyRedirects, ConnectionError, ConnectTimeout, ReadTimeout
from urllib3.exceptions import NewConnectionError
from TheDigger_src.utils.exceptions import RequestHandlerException
from TheDigger_src.utils.singleton import Singleton


class RequestHandler(metaclass=Singleton):
    """
    A wrapper for request sending and session creating.
    Used to abstract proxy/tor routing to avoid repeating configurations for each module
    """
    def __init__(self,
                 proxy_list=None,
                 tor_routing=False,
                 single_proxy=None,
                 delay=None,
                 cookies=None):
        self.proxy_list = proxy_list
        self.tor_routing = tor_routing
        self.delay = delay
        self.single_proxy = single_proxy
        self.proxies = self._set_instance_proxies()
        self.cookies = cookies
        self.allowed_methods = {"GET", "HEAD", "POST"}
        self.headers = self._set_headers()

    @staticmethod
    def _set_headers():
        headers = requests_utils.default_headers()
        headers["User-Agent"] = UserAgent(verify_ssl=False).random
        return headers

    def _set_instance_proxies(self):
        """
        Set the proxies to any of the following:
        Proxy List - a list of proxies to choose randomly from for each request. Read from file.
        TOR - a dict of socks5 and the TOR service default 9050 that will be used
        Else, No proxies - an empty dict will be used.
        """
        proxies = {}
        if self.tor_routing:
            proxies = {
                "http": "socks5://127.0.0.1:9050",
                "https": "socks5://127.0.0.1:9050"
            }
        elif self.proxy_list:
            try:
                with open(self.proxy_list, "r") as file:
                    file = file.readlines()
                    proxies = [x.replace("\n", "") for x in file]
            except FileNotFoundError:
                raise RequestHandlerException("Cannot read proxies from {}".format(self.proxy_list))
        elif self.single_proxy:
            proxies = {
                "http": self.single_proxy,
                "https": self.single_proxy
            }
        return proxies

    def _get_request_proxies(self):
        if self.tor_routing or self.single_proxy:
            proxies = self.proxies
        elif self.proxy_list:
            if not self.proxies:
                raise RequestHandlerException("No valid proxies left in proxy list. Exiting.")
            else:
                try:
                    prx = random.choice(self.proxies)
                    proxies = {proto: "{}://{}".format(proto, prx) for proto in ("http", "https")}
                except IndexError:
                    raise RequestHandlerException("No valid proxies left in proxy list. Exiting.")
        else:
            proxies = self.proxies
        return proxies

    def send(self, method="GET", *args, **kwargs):
        """
        Send a GET/POST/HEAD request using the object's proxies and headers
        :param method: Method to send request in. GET/POST/HEAD
        """
        proxies = self._get_request_proxies()

        try:
            if method.upper() in self.allowed_methods:
                kwargs['timeout'] = kwargs['timeout'] if 'timeout' in kwargs else 5
                return request(method, proxies=proxies, headers=self.headers, cookies=self.cookies, *args, **kwargs)
            else:
                raise RequestHandlerException("Unsupported method: {}".format(method))
        except ProxyError:
            # TODO: Apply fail over for bad proxies or drop them
            raise RequestHandlerException("Error connecting to proxy")
        except (ConnectTimeout, ReadTimeout):
            raise RequestHandlerException("Connection with server timed out")
        except NewConnectionError:
            raise RequestHandlerException("Address cannot be resolved")
            # New connection error == Can't resolve address
        except ConnectionError:
            # TODO: Increase delay
            raise RequestHandlerException("Error connecting to host")
        except TooManyRedirects:
            raise RequestHandlerException("Infinite redirects detected - too many redirects error")
        except UnicodeDecodeError:
            # Following issue #19, apparently some sites do not use utf-8 in their uris :<>
            pass

    def get_new_session(self):
        """Returns a new session using the object's proxies and headers"""
        session = Session()
        session.headers = self.headers
        session.proxies = self._get_request_proxies()
        return session



import requests
from bs4 import BeautifulSoup
from requests.exceptions import ConnectionError, TooManyRedirects
from TheDigger_src.utils.web_server_validator import WebServerValidator
from TheDigger_src.lib.storage_explorer import StorageExplorer
from TheDigger_src.utils.request_handler import RequestHandler
from TheDigger_src.utils.help_utils import HelpUtilities
from TheDigger_src.utils.coloring import COLOR, COLORED_COMBOS
from TheDigger_src.utils.exceptions import WebAppScannerException, WebServerValidatorException
from TheDigger_src.utils.logger import Logger


class WebApplicationScanner:

    def __init__(self, host):
        self.host = host
        self.request_handler = RequestHandler()
        self.web_server_validator = WebServerValidator()
        self.headers = None
        self.robots = None
        self.forms = None
        self.fuzzable_urls = set()
        self.emails = set()
        log_file = HelpUtilities.get_output_path("{}/web_scan.txt".format(self.host.target))
        self.target_dir = "/".join(log_file.split("/")[:-1])
        self.logger = Logger(log_file)
        self.storage_explorer = StorageExplorer(host, self.logger)

    def _detect_cms(self, tries=0):
        """
        Detect CMS using whatcms.org.
        Has a re-try mechanism because false negatives may occur
        :param tries: Count of tries for CMS discovery
        """
        # WhatCMS is under CloudFlare which detects and blocks proxied/Tor traffic, hence normal request.
        page = requests.get(url="https://whatcms.org/?s={}".format(self.host.target))
        soup = BeautifulSoup(page.text, "lxml")
        found = soup.select(".panel.panel-success")
        if found:
            try:
                cms = [a for a in soup.select("a") if "/c/" in a.get("href")][0]
                self.logger.info("{} CMS detected: target is using {}{}{}".format(
                    COLORED_COMBOS.GOOD, COLOR.GREEN, cms.get("title"), COLOR.RESET))
            except IndexError:
                if tries >= 4:
                    return
                else:
                    self._detect_cms(tries=tries + 1)
        else:
            if tries >= 4:
                return
            else:
                self._detect_cms(tries=tries + 1)

    def _cookie_info(self, jar):
        for cookie in jar:
            key = cookie.__dict__.get("name")
            domain = cookie.__dict__.get("domain")
            secure = cookie.__dict__.get("secure")
            http_only = cookie.has_nonstandard_attr("HttpOnly")
            try:
                if domain in self.host.target or self.host.target in domain:
                    if not secure or not http_only:
                        current = "%s Cookie: {%s} -" % (COLORED_COMBOS.GOOD, key)
                        if not secure and not http_only:
                            current += " both secure and HttpOnly flags are not set"
                        elif not secure:
                            current += " secure flag not set"
                        else:
                            current += " HttpOnly flag not set"
                        self.logger.info(current)

            except TypeError:
                continue

    def _server_info(self):
        if self.headers.get("server"):
            self.logger.info("{} Web server detected: {}{}{}".format(
                COLORED_COMBOS.GOOD, COLOR.GREEN, self.headers.get("server"), COLOR.RESET))

    def _x_powered_by(self):
        if self.headers.get("X-Powered-By"):
            self.logger.info("{} X-Powered-By header detected: {}{}{}".format(
                COLORED_COMBOS.GOOD, COLOR.GREEN, self.headers.get("X-Powered-By"), COLOR.RESET))

    def _anti_clickjacking(self):
        if not self.headers.get("X-Frame-Options"):
            self.logger.info(
                "{} X-Frame-Options header not detected - target might be vulnerable to clickjacking".format(
                    COLORED_COMBOS.GOOD)
            )

    def _xss_protection(self):
        xss_header = self.headers.get("X-XSS-PROTECTION")
        if xss_header and "1" in xss_header:
            self.logger.info("{} Found X-XSS-PROTECTION header".format(COLORED_COMBOS.BAD))

    def _cors_wildcard(self):
        if self.headers.get("Access-Control-Allow-Origin") == "*":
            self.logger.info("{} CORS wildcard detected".format(COLORED_COMBOS.GOOD))

    def _robots(self):
        res = self.request_handler.send(
            "GET",
            url="{}://{}:{}/robots.txt".format(
                self.host.protocol,
                self.host.target,
                self.host.port
            )
        )
        if res.status_code != 404 and res.text and "<!DOCTYPE html>" not in res.text:
            self.logger.info("{} Found robots.txt".format(COLORED_COMBOS.GOOD))
            with open("{}/robots.txt".format(self.target_dir), "w") as file:
                file.write(res.text)

    def _sitemap(self):
        res = self.request_handler.send(
            "GET",
            url="{}://{}:{}/sitemap.xml".format(
                self.host.protocol,
                self.host.target,
                self.host.port
            )
        )
        if res.status_code != 404 and res.text and "<!DOCTYPE html>" not in res.text:
            self.logger.info("{} Found sitemap.xml".format(COLORED_COMBOS.GOOD))
            with open("{}/sitemap.xml".format(self.target_dir), "w") as file:
                file.write(res.text)

    def _analyze_hrefs(self, href):
        if all(("?" in href, "=" in href, not href.startswith("mailto:"))):
            if any(((self.host.naked and self.host.naked in href), self.host.target in href, href.startswith("/"))):
                self.fuzzable_urls.add(href)
        elif href.startswith("mailto:"):
            self._add_to_emails(href)

    def _log_fuzzable_urls(self):
        base_target = "{}://{}:{}".format(self.host.protocol, self.host.target, self.host.port)
        for url in self.fuzzable_urls:
            if url.startswith("/"):
                self.logger.debug("\t{}{}".format(base_target, url))
            else:
                self.logger.debug("\t{}".format(url))

    def _log_emails(self):
        for email in self.emails:
            self.logger.debug("\t{}".format(email[7:]))

    def _find_urls(self, soup):
        urls = soup.select("a")
        if urls:
            for url in urls:
                href = url.get("href")
                if href:
                    self._analyze_hrefs(href)

            if self.fuzzable_urls:
                self.logger.info("{} {} fuzzable URLs discovered".format(
                    COLORED_COMBOS.NOTIFY, len(self.fuzzable_urls)))
                self._log_fuzzable_urls()

            if self.emails:
                self.logger.info("{} {} email addresses discovered".format(
                    COLORED_COMBOS.NOTIFY, len(self.emails)))
                self._log_emails()

    def _find_forms(self, soup):
        # TODO: Analyze interesting input names/ids/params
        self.forms = soup.select("form")
        if self.forms:
            self.logger.info("{} {} HTML forms discovered".format(COLORED_COMBOS.NOTIFY, len(self.forms)))
            for form in self.forms:
                form_action = form.get("action")
                if form_action == "#":
                    continue
                form_id = form.get("id")
                form_class = form.get("class")
                form_method = form.get("method")
                self.logger.debug("\tForm details: ID: {}, Class: {}, Method: {}, action: {}".format(
                    form_id, form_class, form_method, form_action
                ))

    def _add_to_emails(self, href):
        self.emails.add(href)

    async def get_web_application_info(self):
        session = self.request_handler.get_new_session()
        try:
            with session:
                # Test if target is serving HTTP requests
                response = session.get(
                    timeout=20,
                    url="{}://{}:{}".format(
                        self.host.protocol,
                        self.host.target,
                        self.host.port
                    )
                )
                self.headers = response.headers
                self._detect_cms()
                self._robots()
                self._sitemap()
                self._server_info()
                self._x_powered_by()
                self._cors_wildcard()
                self._xss_protection()
                self._anti_clickjacking()
                self._cookie_info(session.cookies)

                soup = BeautifulSoup(response.text, "lxml")
                self._find_urls(soup)
                self._find_forms(soup)
                self.storage_explorer.run(soup)

        except (ConnectionError, TooManyRedirects) as e:
            raise WebAppScannerException("Couldn't get response from server.\n"
                                         "Caused due to exception: {}".format(str(e)))

    async def run_scan(self):
        self.logger.info("{} Trying to collect {} web application data".format(COLORED_COMBOS.INFO, self.host))
        try:
            self.web_server_validator.validate_target_webserver(self.host)
            await self.get_web_application_info()
        except WebServerValidatorException:
            self.logger.info(
                "{} Target does not seem to have an active web server on port: {}. "
                "No web application data will be gathered.".format(COLORED_COMBOS.NOTIFY, self.host.port))
            return



import os
from ipaddress import ip_address
from dns.exception import Timeout
from TheDigger_src.lib.dns_handler import DNSHandler
from TheDigger_src.utils.exceptions import HostHandlerException
from TheDigger_src.utils.help_utils import HelpUtilities
from TheDigger_src.utils.coloring import COLOR, COLORED_COMBOS
from TheDigger_src.utils.logger import Logger, SystemOutLogger


class Host:
    """
    Host parsing, IP to host resolution (and vice verse), etc
    Sets domain/IP, port, protocol. also tries to parse FQDN, naked domain, if possible.
    """
    def __init__(self, target, dns_records):
        self.target = target.strip()
        self.dns_records = dns_records
        self.port = 80
        self.protocol = "http"
        self.is_ip = False
        self.fqdn = None
        self.naked = None
        self.dns_results = {}
        self.logger = SystemOutLogger()

    def __str__(self):
        return self.target

    def __repr__(self):
        return self.__dict__

    @staticmethod
    def _create_host_dir(path):
        try:
            os.makedirs("/".join(path.split("/")[:-1]), exist_ok=True)
        except FileExistsError:
            pass

    def validate_ip(self, addr=None):
        if not addr:
            addr = self.target
        try:
            ip_address(addr.strip())
            return True
        except ValueError:
            return

    def _extract_port(self, addr):
        try:
            self.target, self.port = addr.split(":")
            try:
                self.port = int(self.port)
            except ValueError:
                # Probably has a path after the port, e.g - localhost:3000/home.asp
                raise HostHandlerException("Failed to parse port {}. Is there a path after it ?".format(
                    self.port
                ))
            self.logger.info("{} Port detected: {}".format(COLORED_COMBOS.NOTIFY, self.port))
        except IndexError:
            self.logger.info("{} Did not detect port. Using default port 80".format(COLORED_COMBOS.NOTIFY))
            return
        return

    def _is_proto(self, domain=None):
        if not domain:
            domain = self.target
        if "://" in domain:
            if any(domain.startswith(proto) for proto in ("https", "http")):
                return True
            else:
                raise HostHandlerException("Unknown or unsupported protocol: {}".format(self.target.split("://")[0]))
        return

    def write_up(self):
        self.logger.info("{} Writing DNS query results".format(COLORED_COMBOS.GOOD, self))

        for record in self.dns_results:
            self.logger.debug(record+"\n")
            for value in self.dns_results.get(record):
                self.logger.debug("\t{}".format(value))

    def create_host_dir_and_set_file_logger(self):
        log_file = HelpUtilities.get_output_path("{}/dns_records.txt".format(self.target))
        self._create_host_dir(log_file)
        self.logger = Logger(log_file)

    def parse(self):
        """
        Try to extract domain (full, naked, sub-domain), IP and port.
        """
        if self.target.endswith("/"):
            self.target = self.target[:-1]

        if self._is_proto(self.target):
            try:
                self.protocol, self.target = self.target.split("://")
                self.logger.info("{} Protocol detected: {}".format(COLORED_COMBOS.NOTIFY, self.protocol))
                if self.protocol.lower() == "https" and self.port == 80:
                    self.port = 443
            except ValueError:
                raise HostHandlerException("Could not make domain and protocol from host")

        if ":" in self.target:
            self._extract_port(self.target)

        if self.validate_ip(self.target):
            self.logger.info("{} Detected {} as an IP address.".format(COLORED_COMBOS.NOTIFY, self.target))
            self.is_ip = True
        else:
            domains = []
            if self.target.startswith("www."):
                # Obviously an FQDN
                domains.extend((self.target, self.target.split("www.")[1]))
                self.fqdn = self.target
                self.naked = ".".join(self.fqdn.split('.')[1:])
            else:
                domains.append(self.target)
                domain_levels = self.target.split(".")
                if len(domain_levels) == 2 or (len(domain_levels) == 3 and domain_levels[1] == "co"):
                    self.logger.info("{} Found {} to be a naked domain".format(COLORED_COMBOS.NOTIFY, self.target))
                    self.naked = self.target

            try:
                self.dns_results = DNSHandler.query_dns(domains, self.dns_records)
            except Timeout:
                raise HostHandlerException("DNS Query timed out. Maybe target has DNS protection ?")

            if self.dns_results.get("CNAME"):
                # Naked domains shouldn't hold CNAME records according to RFC regulations
                self.logger.info("{} Found {} to be an FQDN by CNAME presence in DNS records".format(
                    COLORED_COMBOS.NOTIFY, self.target))

                self.fqdn = self.target
                self.naked = ".".join(self.fqdn.split('.')[1:])
        self.create_host_dir_and_set_file_logger()
        self.write_up()



import logging
from os import path
from sys import stdout
from TheDigger_src.utils.singleton import Singleton


class SystemOutLogger(metaclass=Singleton):
    """
    Single instance stdout logger to be shared among modules
    Logging level is set by verbosity/quiet arguments from user
    Logs to stdout - other loggers call its functions to log to stdout
    in addition to their own file-writing logging
    """
    def __init__(self, level="INFO"):
        self.level = level
        self.logger = self.get_logger()

    def get_logger(self):
        logger = logging.getLogger("TheDigger")
        logger.setLevel(self.level)

        out_handler = logging.StreamHandler(stdout)
        formatter = logging.Formatter('%(message)s')
        out_handler.setFormatter(formatter)
        logger.addHandler(out_handler)
        return logger

    def debug(self, *args, **kwargs):
        self.logger.debug(*args, **kwargs)

    def info(self, *args, **kwargs):
        self.logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs):
        self.logger.warning(*args, **kwargs)

    def error(self, *args, **kwargs):
        self.logger.error(*args, **kwargs)

    def critical(self, *args, **kwargs):
        self.logger.critical(*args, **kwargs)


class Logger:
    """
    Logger that should instantiated for each module
    Will write all logs (DEBUG) to self.outfile argument.
    In addition calls SystemOutLogger functions to write to stdout in correspondence with
    verbosity levels
    """

    def __init__(self, outfile):
        self.outfile = outfile
        self.stout_logger = SystemOutLogger()
        self.logger = self.get_logger()

    def get_logger(self):
        logger = logging.getLogger(self.__str__())
        logger.setLevel("DEBUG")

        out_handler = logging.FileHandler(self.outfile)
        formatter = logging.Formatter('%(message)s')
        out_handler.setFormatter(formatter)
        logger.addHandler(out_handler)
        return logger

    def debug(self, *args, **kwargs):
        self.stout_logger.debug(*args, **kwargs)
        self.logger.debug(*args, **kwargs)

    def info(self, *args, **kwargs):
        self.stout_logger.info(*args, **kwargs)
        self.logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs):
        self.stout_logger.warning(*args, **kwargs)
        self.logger.warning(*args, **kwargs)

    def error(self, *args, **kwargs):
        self.stout_logger.error(*args, **kwargs)
        self.logger.error(*args, **kwargs)

    def critical(self, *args, **kwargs):
        self.stout_logger.critical(*args, **kwargs)
        self.logger.critical(*args, **kwargs)



import os
import xmltodict
from TheDigger_src.utils.request_handler import RequestHandler
from TheDigger_src.utils.exceptions import TheDiggerException, RequestHandlerException
from TheDigger_src.utils.coloring import COLORED_COMBOS, COLOR


# Set path for relative access to builtin files.
MY_PATH = os.path.abspath(os.path.dirname(__file__))
HTTP = "http://"
HTTPS = "https://"
BASE_S3_URL = "s3.amazonaws.com"


class Storage:

    def __init__(self, host, logger):
        self.host = host
        self.logger = logger
        self.request_handler = RequestHandler()
        self.storage_urls_found = set()
        self.num_files_found = 0
        file_list_path = os.path.join(MY_PATH, "../wordlists/storage_sensitive")
        with open(file_list_path, "r") as file:
            files = file.readlines()
            self.sensitive_files = [x.replace("\n", "") for x in files]

    @staticmethod
    def _normalize_url(url):
        if url.startswith(HTTP):
            url = url.replace(HTTP, "")
            url = "".join([part for part in url.split("//") if part])
            return HTTP+url
        else:
            url = url.replace(HTTPS, "")
            url = "".join([part for part in url.split("//") if part])
            return HTTPS+url


# Is this a thing ??
class AzureStorageHandler:
    pass


class GoogleStorageHandler:
    pass


class AmazonS3Handler(Storage):

    def __init__(self, host, logger):
        super().__init__(host, logger)
        self.s3_buckets = set()

    def _is_s3_url(self, src):
        # Not including third party Amazon host services - aka cdn.3rdparty.com
        return any(("s3" in src and "amazonaws" in src,
                    "cdn.{}".format(str(self.host.naked)) in src,
                    "cdn.{}".format(self.host.target) in src,
                    "cdn.{}".format(".".join(self.host.target.split(".")[1:])) in src,
                    "cloudfront.net" in src))

    @staticmethod
    def _is_amazon_s3_bucket(res):
        return res.headers.get("Server") == "AmazonS3"

    def _test_s3_bucket_permissions(self, bucket):
        try:
            bucket_url = [part for part in bucket.no_scheme_url.split("/") if part]
            bucket_len = len(bucket_url)

            for i in range(bucket_len-1):
                url = "/".join(bucket_url[:i+1])
                if url == BASE_S3_URL or url in self.storage_urls_found:
                    continue

                self.storage_urls_found.add(url)
                res = self.request_handler.send("GET", url=HTTPS+url)

                if res.status_code == 200 and res.headers.get("Content-Type") == "application/xml":
                    self.logger.info("{} Vulnerable S3 bucket detected: {}{}{}. Enumerating sensitive files".format(
                        COLORED_COMBOS.GOOD, COLOR.RED, url, COLOR.RESET))
                    bucket.vulnerable = True
                    self._scan_for_sensitive_files(res.text, url)

        except RequestHandlerException:
            # Cannot connect to bucket, move on
            pass

    def _scan_for_sensitive_files(self, contents, url):
        xpars = xmltodict.parse(contents)
        for el in xpars.get("ListBucketResult").get("Contents"):
            key = el.get("Key")
            for file in self.sensitive_files:
                if file in key:
                    self.logger.debug("Found {} file in bucket {}".format(key, url))
                    self.num_files_found += 1


class S3Bucket:

    def __init__(self, url):
        self.url = self._strip_resource_from_bucket(url)
        self.no_scheme_url = self._remove_scheme_from_url(self.url)
        self.vulnerable = False

    @staticmethod
    def _strip_resource_from_bucket(bucket_url):
        # Return the storage URL without the resource
        return "/".join(bucket_url.split("/")[:-1])

    @staticmethod
    def _remove_scheme_from_url(url):
        if url.startswith(HTTP):
            url = url.replace(HTTP, "")
        else:
            url = url.replace(HTTPS, "")
        return "".join([part for part in url.split("//") if part])


class StorageExplorer(AmazonS3Handler, GoogleStorageHandler, AzureStorageHandler):
    """
    Find and test privileges of target cloud storage and look for sensitive files in it.
    Can lead to finding .git/.DS_Store/etc files with tokens, passwords and more.
    """

    def __init__(self, host, logger):
        super().__init__(host, logger)
        self.host = host
        self.logger = logger  # Uses the logger from web_app module
        self.buckets_found = set()

    @staticmethod
    def _get_image_sources_from_html(soup):
        images = soup.select("img")
        return {img.get("src") for img in images if img.get("src")}

    def _add_to_found_storage(self, storage_url):
        """
        Will first normalize the img src and then check if this bucket was discovered before
        If it is in storage_urls_found, the function returns
        Else, it send a GET for the original URL (normalized image src) and will look for "AmazonS3" in
        the "Server" response header.
        If found, will add to URL with the resource stripped

        :param storage_url: img src scraped from page
        """
        storage_url = self._normalize_url(storage_url)
        bucket = S3Bucket(storage_url)
        if bucket.url not in self.storage_urls_found:
            try:
                res = self.request_handler.send("GET", url=storage_url)
                if self._is_amazon_s3_bucket(res):
                    self.storage_urls_found.add(bucket.url)
                    self.s3_buckets.add(bucket)
            except RequestHandlerException:
                # Cannot connect to storage, move on
                pass

    def run(self, soup):
        img_srcs = self._get_image_sources_from_html(soup)
        # First validation
        urls = {src for src in img_srcs if self._is_s3_url(src)}
        for url in urls:
            self._add_to_found_storage(url)
        if self.s3_buckets:
            self.logger.info("{} S3 buckets discovered. Testing for permissions".format(COLORED_COMBOS.NOTIFY))
            for bucket in self.s3_buckets:
                if bucket.no_scheme_url in self.storage_urls_found:
                    continue
                else:
                    self._test_s3_bucket_permissions(bucket)

            if self.num_files_found > 0:
                self.logger.info(
                    "{} Found {}{}{} sensitive files in S3 buckets. inspect web scan logs for more information.".format(
                        COLORED_COMBOS.GOOD, COLOR.GREEN, self.num_files_found, COLOR.RESET))
            elif any(b.vulnerable for b in self.s3_buckets):
                self.logger.info("{} No sensitive files found in target's cloud storage".format(COLORED_COMBOS.BAD))
            else:
                self.logger.info("{} Could not access target's cloud storage."
                                 " All permissions are set properly".format(COLORED_COMBOS.BAD))



import re
from subprocess import PIPE, Popen
from TheDigger_src.utils.help_utils import HelpUtilities
from TheDigger_src.utils.logger import Logger
from TheDigger_src.utils.coloring import COLOR, COLORED_COMBOS


class NmapScan:
    """
    Nmap scan class
    Will run SYN/TCP scan according to privileges.
    By default TheDigger will run -sT run TheDigger with sudo for -sS
    """

    def __init__(self, host, port_range, full_scan=None, scripts=None, services=None):
        self.target = host.target
        self.full_scan = full_scan
        self.scripts = scripts
        self.services = services
        self.port_range = port_range
        self.path = HelpUtilities.get_output_path("{}/nmap_scan.txt".format(self.target))
        self.logger = Logger(self.path)

    def build_script(self):
        script = ["nmap", "-Pn", self.target]

        if self.port_range:
            HelpUtilities.validate_port_range(self.port_range)
            script.append("-p")
            script.append(self.port_range)
            self.logger.info("{} Added port range {} to Nmap script".format(COLORED_COMBOS.NOTIFY, self.port_range))
        if self.full_scan:
            script.append("-sV")
            script.append("-sC")
            self.logger.info("{} Added scripts and services to Nmap script".format(COLORED_COMBOS.NOTIFY))
            return script
        else:
            if self.scripts:
                self.logger.info("{} Added safe-scripts scan to Nmap script".format(COLORED_COMBOS.NOTIFY))
                script.append("-sC")
            if self.services:
                self.logger.info("{} Added service scan to Nmap script".format(COLORED_COMBOS.NOTIFY))
                script.append("-sV")
        return script


class NmapVulnersScan(NmapScan):
    """
    NmapVulners scan class (NmapScan subclass)
    """

    def __init__(self, host, port_range, vulners_path):
        super().__init__(host=host, port_range=port_range)
        self.vulners_path = vulners_path
        self.path = HelpUtilities.get_output_path("{}/nmap_vulners_scan.txt".format(self.target))
        self.logger = Logger(self.path)

    def build_script(self):
        script = ["nmap", "-Pn", "-sV", "--script", self.vulners_path, self.target]

        if self.port_range:
            HelpUtilities.validate_port_range(self.port_range)
            script.append("-p")
            script.append(self.port_range)
            self.logger.info("{} Added port range {} to Nmap script".format(COLORED_COMBOS.NOTIFY, self.port_range))

        return script


class Scanner:

    @classmethod
    def run(cls, scan):
        script = scan.build_script()

        scan.logger.info("{} Nmap script to run: {}".format(COLORED_COMBOS.INFO, " ".join(script)))
        scan.logger.info("{} Nmap scan started\n".format(COLORED_COMBOS.GOOD))
        process = Popen(
            script,
            stdout=PIPE,
            stderr=PIPE
        )
        result, err = process.communicate()
        result, err = result.decode().strip(), err.decode().strip()
        if result:
            parsed_result = cls._parse_scan_output(result)
            scan.logger.info(parsed_result)
        Scanner.write_up(scan, result, err)

    @classmethod
    def _parse_scan_output(cls, result):
        parsed_output = ""
        for line in result.split("\n"):
            if "PORT" in line and "STATE" in line:
                parsed_output += "{} Nmap discovered the following ports:\n".format(COLORED_COMBOS.GOOD)
            if "/tcp" in line or "/udp" in line and "open" in line:
                line = line.split()
                parsed_output += "\t{}{}{} {}\n".format(COLOR.GREEN, line[0], COLOR.RESET, " ".join(line[1:]))
        return parsed_output

    @classmethod
    def write_up(cls, scan, result, err):
        open(scan.path, "w").close()
        if result:
            scan.logger.debug(result+"\n")
        if err:
            scan.logger.debug(err)


class VulnersScanner(Scanner):

    @classmethod
    def _parse_scan_output(cls, result):

        parsed_output = ""
        out_versions, out_pure = cls._parse_vulners_output(result)

        out_versions = re.sub(r"(\d+\/(?:tcp|udp))", COLOR.GREEN + r"\1" + COLOR.RESET, out_versions)
        out_versions = re.sub(r"(\sCVE\S*)", COLOR.RED + r"\1" + COLOR.RESET, out_versions)
        out_pure = re.sub(r"(\d+\/(?:tcp|udp))", COLOR.GREEN + r"\1" + COLOR.RESET, out_pure)

        if out_pure:
            parsed_output += "{} NmapVulners discovered the following open ports:\n{}"\
                .format(COLORED_COMBOS.GOOD, out_pure)
        if out_versions:
            parsed_output += "{} NmapVulners discovered some vulnerable software within the following open ports:\n{}"\
                .format(COLORED_COMBOS.GOOD, out_versions)
        return parsed_output

    @classmethod
    def _parse_vulners_output(cls, res):
        ports = re.findall(r"(?:^\d+/(?:tcp|udp).*open.*$\n(?:^\|.*$\n)*)", res, re.MULTILINE)
        out_vers = ""
        out_none = ""
        for port in ports:
            if 'vulners' in port:
                out_vers += '\n' + '\n'.join(
                    re.findall(r"^(\d+/(?:tcp|udp).*open.*$)[\s\S]*?(^\|.*vulners[\s\S]+?^\|_.+?$)", port,
                               re.MULTILINE)[0])
            else:
                out_none += port
        return out_vers, out_none



import re
from platform import system
# noinspection PyProtectedMember
from asyncio.subprocess import PIPE, create_subprocess_exec
from TheDigger_src.utils.help_utils import HelpUtilities
from TheDigger_src.utils.coloring import COLOR, COLORED_COMBOS
from TheDigger_src.utils.logger import Logger


# macOS support
if system() == "Darwin":
    TIMEOUT = "gtimeout"
else:
    TIMEOUT = "timeout"


class TLSCipherSuiteChecker:

    def __init__(self, host):
        self.target = host.target

    async def scan_ciphers(self, port):
        script = "nmap --script ssl-enum-ciphers -p {} {}".format(str(port), self.target).split()
        process = await create_subprocess_exec(
            *script,
            stdout=PIPE,
            stderr=PIPE
        )
        result, err = await process.communicate()
        if process.returncode != 0:
            parsed = err.decode().strip()
        else:
            parsed = self._parse_cipher_scan_outpt(result)
        return parsed

    @staticmethod
    def _color_warnings_and_weak_ciphers(result):
        for index, line in enumerate(result):
            if line.endswith("- C") or line.endswith("- D") or line.endswith("- E"):
                colored = line + " - {}WEAK{}".format(COLOR.RED, COLOR.RESET)
                result.insert(index, colored)
                result.pop(index+1)
            elif "warnings:" in line:
                curr = index+1
                while "TLSv" not in result[curr] and "least strength" not in result[curr]:
                    colored = "{}{}{}".format(COLOR.RED, result[curr], COLOR.RESET)
                    result.insert(curr, colored)
                    result.pop(curr+1)
                    curr += 1
        return result[1:]

    def _parse_cipher_scan_outpt(self, result):
        result = [line for line in result.decode().strip().split("\n") if "|" in line]
        result = self._color_warnings_and_weak_ciphers(result)
        return '\n'.join(result)


# noinspection PyTypeChecker
class TLSHandler(TLSCipherSuiteChecker):

    def __init__(self, host, port):
        super().__init__(host)
        self.target = host.target
        self.port = port
        self._versions = ("tls1", "tls1_1", "tls1_2")
        # OpenSSL likes to hang, hence the timeout call
        self._base_script = "{} 10 openssl s_client -connect {}:{} ".format(TIMEOUT, self.target, self.port)
        self.begin = "-----BEGIN CERTIFICATE-----"
        self.end = "-----END CERTIFICATE-----"
        self.sni_data = {}
        self.non_sni_data = {}
        self.ciphers = ""
        log_file = HelpUtilities.get_output_path("{}/tls_report.txt".format(self.target))
        self.logger = Logger(log_file)

    def _tls_results_exist(self):
        if all(not x for x in (self.ciphers, *self.non_sni_data.values(), *self.sni_data.values())):
            return
        return True

    def _are_certificates_identical(self):
        """
        Validate that both certificates exist.
        :returns: True if they are identical, False otherwise
        """
        sni_cert = self.sni_data.get("Certificate_details")
        non_sni_cert = self.non_sni_data.get("Certificate_details")
        if all(cert for cert in (sni_cert, non_sni_cert) if cert) and sni_cert == non_sni_cert:
            return True
        return

    def _is_certificate_exists(self, text):
        if self.begin in text and self.end in text:
            return True
        return

    async def _extract_certificate_details(self, data):
        process = await create_subprocess_exec(
            TIMEOUT, "5", "openssl", "x509", "-text",
            stdin=PIPE,
            stderr=PIPE,
            stdout=PIPE
        )
        result, err = await process.communicate(input=bytes(data, encoding='ascii'))
        result = result.decode().strip()
        cert_details = result.split(self.begin)[0].strip()

        result_lines = cert_details.split("\n")
        for i, line in enumerate(result_lines):
            if "DNS:" in line:
                result_lines.pop(i)
                result_lines.pop(i-1)

        cert_details = "\n".join(result_lines)
        return cert_details

    async def _is_heartbleed_vulnerable(self):
        script = self._base_script + "-tlsextdebug"
        process = await create_subprocess_exec(
            *script.split(),
            stdout=PIPE,
            stderr=PIPE
        )
        result, err = await process.communicate()
        try:
            if "server extension \"heartbeat\" (id=15)" in result.decode().strip():
                self.logger.info("{} Target seems to be vulnerable to Heartbleed - CVE-2014-016. "
                                 "see http://heartbleed.com/ for more details.".format(COLORED_COMBOS.GOOD))
        except TypeError:  # Type error means no result
            pass

    async def _execute_ssl_data_extraction(self, sni=False):
        """
        Test for version support (SNI/non-SNI), get all SANs, get certificate details
        :param sni: True will call cause _exec_openssl to call openssl with -servername flag
        """
        # Do for all responses
        responses = await self._run_openssl_sclient_cmd(self._base_script, sni)
        tls_dict = self._parse_openssl_sclient_output(responses)
        # Do for one successful SSL response
        for res in responses:
            if self._is_certificate_exists(res):
                tls_dict["SANs"] = await self._get_sans_from_openssl_cmd(res)
                tls_dict["Certificate_details"] = await self._extract_certificate_details(res)
                break

        return tls_dict

    async def _run_openssl_sclient_cmd(self, script, sni=False):
        processes = []
        outputs = []
        if sni:
            script += " -servername {}".format(self.target)
        for v in self._versions:
            curr = (script + ' -{}'.format(v)).split()
            processes.append(
                await create_subprocess_exec(
                    *curr,
                    stdout=PIPE,
                    stderr=PIPE
                )
            )
        for p in processes:
            result, err = await p.communicate()

            outputs.append(result.decode().strip())

        return outputs

    @staticmethod
    async def _get_sans_from_openssl_cmd(data):
        process = await create_subprocess_exec(
            "openssl", "x509", "-noout", "-text",
            stdin=PIPE,
            stderr=PIPE,
            stdout=PIPE
        )
        result, err = await process.communicate(input=bytes(data, encoding='ascii'))
        sans = re.findall(r"DNS:\S*\b", result.decode().strip())
        return {san.replace("DNS:", '') for san in sans}

    def _parse_openssl_sclient_output(self, results):
        is_supported = {"TLSv1": False, "TLSv1.1": False, "TLSv1.2": False}
        for res in results:
            if not self._is_certificate_exists(res):
                continue
            for line in res.split('\n'):
                if "Protocol" in line:
                    ver = line.strip().split(':')[1].strip()
                    is_supported[ver] = True
        return is_supported

    def _dictionary_log_procedure(self, result_dict):
        for k, v in result_dict.items():
            if k == "SANs":
                self.logger.debug("{0}:\n{1}\n {2}\n{1}\n".format(k, "-"*15, "\n".join(v)))
            elif k == "Certificate_details":
                self.logger.debug(v)
            else:
                self.logger.debug("{}: {}\n".format(k, v))

    def write_up(self):
        self.logger.info("{} Supported Ciphers:".format(COLORED_COMBOS.GOOD))
        self.logger.info(self.ciphers+"\n")
        self.logger.debug("-"*80+"\n")
        self.logger.debug("SNI Data:\n")
        self._dictionary_log_procedure(self.sni_data)
        self.logger.debug("-"*80+"\n")
        self.logger.debug("non-SNI Data:\n")
        self._dictionary_log_procedure(self.non_sni_data)

    async def run(self):
        self.logger.info("{} Started collecting TLS data for {}".format(COLORED_COMBOS.INFO, self.target))
        self.ciphers = await self.scan_ciphers(self.port)
        self.non_sni_data = await self._execute_ssl_data_extraction()
        self.sni_data = await self._execute_ssl_data_extraction()
        await self._is_heartbleed_vulnerable()

        if self._tls_results_exist():
            self.logger.info("{} Done collecting TLS data".format(COLORED_COMBOS.INFO))
            if self._are_certificates_identical():
                self.non_sni_data["Certificate_details"] = "Same as SNI Certificate"
            self.write_up()
        else:
            self.logger.info(
                "{} Could not obtain any TLS data from target on port {}. "
                "Target may not support SSL/TLS or supports it on a different port.".format(
                    COLORED_COMBOS.BAD, self.port)
            )



from setuptools import setup, find_packages


with open("README.md", "r") as file:
    long_description = file.read()

setup(
    name='TheDigger',
    packages=find_packages(exclude="tests"),
    license="MIT",
    version='0.1',
    description='Red Team and Offensive Security Scanning and Reconnaisance Tool for Companies and Pentesters. ',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Kartavya Trivedi',
    author_email='Kartavyatrivedi9@gmail.com',
    url='https://github.com/Jistrokz/TheDigger',
    install_requires=['beautifulsoup4',
                      'requests',
                      'dnspython',
                      "lxml",
                      "click",
                      "fake-useragent",
                      "requests[socks]",
                      "xmltodict"],
    package_data={
        "TheDigger_src": [
            "wordlists/*"
        ]
    },
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'TheDigger=TheDigger_src.main:main'
        ]
    },
)



import re
from bs4 import BeautifulSoup
from TheDigger_src.utils.request_handler import RequestHandler
from TheDigger_src.lib.fuzzer import URLFuzzer
from TheDigger_src.utils.help_utils import HelpUtilities
from TheDigger_src.utils.exceptions import TheDiggerException
from TheDigger_src.utils.logger import Logger
from TheDigger_src.utils.coloring import COLOR, COLORED_COMBOS


class SubDomainEnumerator:

    def __init__(self,
                 host,
                 sans,
                 domain_list,
                 ignored_response_codes,
                 num_threads,
                 follow_redirects,
                 no_sub_enum):
        self.host = host
        self.target = host.target
        self.sans = sans
        self.domain_list = domain_list
        self.ignored_error_codes = ignored_response_codes
        self.num_threads = num_threads
        self.follow_redirects = follow_redirects
        self.no_sub_enum = no_sub_enum
        self.request_handler = RequestHandler()
        log_file = HelpUtilities.get_output_path("{}/subdomains.txt".format(self.target))
        self.logger = Logger(log_file)

    async def run(self):
        self.logger.info("{} Enumerating Subdomains".format(COLORED_COMBOS.INFO))
        if self.sans:
            self._extract_from_sans()
        self._google_dork()
        self._extract_from_dns_dumpster()
        if not self.no_sub_enum:
            await self.bruteforce()
        self.logger.info("{} Done enumerating Subdomains".format(COLORED_COMBOS.INFO))

    def _extract_from_sans(self):
        """Looks for different TLDs as well as different sub-domains in SAN list"""
        self.logger.info("{} Trying to find Subdomains in SANs list".format(COLORED_COMBOS.NOTIFY))
        if self.host.naked:
            domain = self.host.naked
            tld_less = domain.split(".")[0]
        else:
            domain = self.host.target.split(".")
            tld_less = domain[1]
            domain = ".".join(domain[1:])

        for san in self.sans:
            if (tld_less in san or domain in san) and self.target != san and not san.startswith("*"):
                self.logger.info("{} Subdomain detected: {}".format(COLORED_COMBOS.GOOD, san))

    def _google_dork(self):
        self.logger.info("{} Trying to discover subdomains in Google".format(COLORED_COMBOS.NOTIFY))
        page = self.request_handler.send(
            "GET",
            url="https://www.google.com/search?q=site:{}&num=100".format(self.target)
        )
        soup = BeautifulSoup(page.text, "lxml")
        results = set(re.findall(r"\w+\.{}".format(self.target), soup.text))
        for subdomain in results:
            if "www." not in subdomain:
                self.logger.info("{} Detected subdomain through Google dorking: {}".format(
                    COLORED_COMBOS.GOOD, subdomain))

    def _extract_from_dns_dumpster(self):
        self.logger.info("{} Trying to extract subdomains from DNS dumpster".format(COLORED_COMBOS.NOTIFY))
        try:
            page = HelpUtilities.query_dns_dumpster(host=self.host)
            soup = BeautifulSoup(page.text, "lxml")
            hosts_table = soup.select(".table")[-1]
            for row in hosts_table.find_all("tr"):
                tds = row.select("td")
                sub_domain = tds[0].text.split('\n')[0]  # Grab just the URL, truncate other information
                self.logger.info("{} Found subdomain in DNS dumpster: {}".format(COLORED_COMBOS.GOOD, sub_domain))
        except (TheDiggerException, IndexError):
            self.logger.info("{} Failed to query DNS dumpster for subdomains".format(COLORED_COMBOS.BAD))

    async def bruteforce(self):
        path = "{}/subdomain_fuzz.txt".format(self.host.target)

        # If a naked domain exists, use it
        if self.host.naked:
            self.host.target = self.host.naked

        self.logger.info("{} Bruteforcing subdomains".format(COLORED_COMBOS.NOTIFY))
        sub_domain_fuzzer = URLFuzzer(
            host=self.host,
            path_to_wordlist=self.domain_list,
            num_threads=self.num_threads,
            ignored_response_codes=self.ignored_error_codes,
            follow_redirects=self.follow_redirects
            )
        await sub_domain_fuzzer.fuzz_all(sub_domain=True, log_file_path=path)



from urllib3.exceptions import HTTPError
from requests.exceptions import ConnectionError, TooManyRedirects
from TheDigger_src.utils.request_handler import RequestHandler
from TheDigger_src.utils.singleton import Singleton
from TheDigger_src.utils.exceptions import WebServerValidatorException, RequestHandlerException


class WebServerValidator(metaclass=Singleton):

    def __init__(self):
        self.request_handler = RequestHandler()

    def validate_target_webserver(self, host):
        try:
            self.request_handler.send(
                "GET",
                timeout=20,
                url="{}://{}:{}".format(
                    host.protocol,
                    host.target,
                    host.port
                )
            )
            return True
        except RequestHandlerException:
            raise WebServerValidatorException
